package checkers.nullness;

import javax.lang.model.element.*;

import javax.lang.model.type.*;

import checkers.basetype.BaseTypeChecker;
import checkers.flow.*;
import checkers.nullness.quals.*;
import checkers.quals.*;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

/**
 * Adds the {@link NonNull} annotation to a type that is:
 *
 * 1. in the scope of a {@link Default} annotation and matches its location
 *    criteria
 * 2. determined to be {@link NonNull} by flow-sensitive inference
 * 3. the class in a static member access (e.g., "System" in "System.out")
 * 4. an array-creation expression (with new)
 * 5. an object-creation expression (with new)
 * 6. a string literal
 * 7. a package declaration
 * 8. an exception parameter
 * 9. the receiver type of a non-static (and non-constructor) method
 *
 * The type factory will also add the {@link Raw} annotation to the receiver
 * type of constructors.
 *
 * <p>
 *
 * Additionally, the type factory will add the {@link Nullable} annotation to a
 * type if the input is the null literal or of type {@link Void}, and may
 * resolve the types of some {@link NonNull} fields as {@link Nullable}
 * depending on the presence of a {@link Raw} annotation on a constructor or
 * method receiver.
 */
public class NullnessAnnotatedTypeFactory extends AnnotatedTypeFactory {

    private final Flow flow;
    private final QualifierDefaults defaults;
    private final TypeAnnotator typeAnnotator;
    private final TreeAnnotator treeAnnotator;
    private final QualifierPolymorphism poly;

    private final AnnotationCompleter completer = new AnnotationCompleter();

    /** Represents {@link NonNull}. */
    final AnnotationMirror NONNULL;

    /** Represents {@link Raw}. */
    final AnnotationMirror RAW;

    /** Represents {@link Nullable}. */
    final AnnotationMirror NULLABLE;

    /** Creates a {@link NullnessAnnotatedTypeFactory}. */
    public NullnessAnnotatedTypeFactory(NullnessChecker checker,
            CompilationUnitTree root) {
        super(checker, root);

        typeAnnotator = new NonNullTypeAnnotator(checker);
        treeAnnotator = new NonNullTreeAnnotator(checker);

        NONNULL = this.annotations.fromClass(NonNull.class);
        RAW = this.annotations.fromClass(Raw.class);
        NULLABLE = this.annotations.fromClass(Nullable.class);

        defaults = new QualifierDefaults(this, this.annotations);

        this.poly = new QualifierPolymorphism(checker, this);

        flow = new NullnessFlow(checker, root, this);
        flow.scan(root, null);
    }

    @Override
    protected void annotateImplicit(Element elt, AnnotatedTypeMirror type) {

        if (elt instanceof VariableElement)
            annotateIfStatic(elt, type);
        typeAnnotator.visit(type);
        defaults.annotate(elt, type);
        poly.annotate(elt, type);
        completer.visit(type);
    }

    @Override
    protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
        treeAnnotator.visit(tree, type);
        typeAnnotator.visit(type);
        defaults.annotate(tree, type);
        substituteRaw(tree, type);
        final AnnotationMirror inferred = flow.test(tree);
        if (inferred != null) {
            type.clearAnnotations();
            type.addAnnotation(inferred);
        }
        poly.annotate(tree, type);
        completer.visit(type);
    }

    /**
     * Substitutes {@link Raw} annotations on a type. If the select expression
     * (which might implicitly be "this") is {@link Raw}, this replaces the
     * annotations on {@code type} with {@link Raw}.
     *
     * @param tree a (possibly implicit) select expression
     * @param type the type for which to substitute annotations based on the
     *        placement of {@link Raw}
     * @return true iff we reduced the annotated type from {@link NonNull} to
     *         {@link Raw}
     */
    boolean substituteRaw(Tree tree, AnnotatedTypeMirror type) {

        // If it's not an expression tree, it's definitely not a select.
        if (!(tree instanceof ExpressionTree))
            return false;

        // Identifiers might implicitly select from "this", but not if they're
        // class or interface identifiers.
        if (tree instanceof IdentifierTree) {

            Element elt = TreeUtils.elementFromUse((IdentifierTree) tree);
            if (elt == null || !elt.getKind().isField())
                return false;

            Tree decl = declarationFromElement(elt);
            if (this.getPath(decl) != null
                        && decl instanceof VariableTree
                        && ((VariableTree) decl).getInitializer() != null
                        && getAnnotatedType(((VariableTree) decl).getInitializer()).hasAnnotation(NONNULL))
            return false;
        }

        final AnnotatedTypeMirror select = getReceiver((ExpressionTree) tree);
        if (select != null && select.hasAnnotation(RAW) && !type.hasAnnotation(NULLABLE)) {
            boolean wasNN = type.hasAnnotation(NONNULL);
            type.clearAnnotations();
            type.addAnnotation(RAW);
            return wasNN;
        }
        return false;
    }

    private class AnnotationCompleter extends AnnotatedTypeScanner<Void, Void> {

        @Override
        protected Void scan(AnnotatedTypeMirror type, Void p) {

            if (type == null)
                return super.scan(type, p);

            if (type.getAnnotations().isEmpty())
                type.addAnnotation(NULLABLE);
            else if (type.hasAnnotation(RAW))
                type.removeAnnotation(NONNULL);
            else if (type.hasAnnotation(NONNULL))
                type.removeAnnotation(NULLABLE);

            if (type instanceof AnnotatedDeclaredType) {
                Name name = TypesUtils.getQualifiedName((DeclaredType)type.getUnderlyingType());
                if (name.contentEquals("java.lang.Void")) {
                    type.clearAnnotations();
                    type.addAnnotation(NULLABLE);
                }
            }

            assert !type.getAnnotations().isEmpty() : type;

            return super.scan(type, p);
        }
    }

    /**
     * Adds {@link NonNull} annotations to a type based on the underlying
     * (unannotated) type.
     */
    private class NonNullTypeAnnotator extends TypeAnnotator {

        /** Creates a {@link NonNullTypeAnnotator} for the given checker. */
        NonNullTypeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Void p) {

            // Don't add implicit receiver annotations if some are already there.
            if (!type.getReceiverType().getAnnotations().isEmpty())
                return super.visitExecutable(type, p);

            ExecutableElement elt = type.getElement();
            assert elt != null;

            // Add @Raw to constructors.
            if (elt.getKind() == ElementKind.CONSTRUCTOR)
                type.getReceiverType().addAnnotation(RAW);

            // Add @NonNull to non-static non-constructors.
            if (!ElementUtils.isStatic(elt))
                type.getReceiverType().addAnnotation(NONNULL);

            return super.visitExecutable(type, p);
        }
    }

    /**
     * Determines whether the element is {@link NonNull} when used in a static
     * member access and modifies the element's type accordingly (by adding
     * {@link NonNull}).
     *
     * @param elt the element being accessed
     * @param type the type of the element {@code elt}
     */
    private void annotateIfStatic(Element elt, AnnotatedTypeMirror type) {

        if (elt == null)
            return;

        if (elt.getKind().isClass() || elt.getKind().isInterface()
                // Workaround for System.{out,in,err} issue: assume all static
                // fields in java.lang.System are nonnull.
                || isSystemField(elt)) {
            type.clearAnnotations();
            type.addAnnotation(NONNULL);
        }
    }

    private boolean isSystemField(Element elt) {
        if (!elt.getKind().isField())
            return false;

        if (!ElementUtils.isStatic(elt) || !ElementUtils.isFinal(elt))
            return false;

        VariableElement var = (VariableElement)elt;

        // Heuristic: if we have a static final field in a system package,
        // treat it as NonNull (many like Boolean.TYPE and System.out
        // have constant value null but are set by the VM).
        boolean inJavaPackage = ElementUtils.getQualifiedClassName(var).toString().startsWith("java.");

        return (var.getConstantValue() != null
                || var.getSimpleName().contentEquals("class")
                || inJavaPackage);
    }

    /**
     * Adds {@link NonNull} annotations to a type based on the AST from which
     * the type was obtained.
     */
    private class NonNullTreeAnnotator extends TreeAnnotator {

        /** Creates a {@link NonNullTreeAnnotator} for the given checker. */
        NonNullTreeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        @Override
        public Void visitMemberSelect(MemberSelectTree node, AnnotatedTypeMirror type) {

            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;

            annotateIfStatic(elt, type);
            return super.visitMemberSelect(node, type);
        }

        @Override
        public Void visitIdentifier(IdentifierTree node, AnnotatedTypeMirror type) {

            Element elt = TreeUtils.elementFromUse(node);
            assert elt != null;

            if (elt.getKind() == ElementKind.PARAMETER
                    && TypesUtils.isThrowable(elt.asType())) {
                // Workaround: exception parameters should be implicitly
                // NonNull, but due to a compiler bug they have
                // kind PARAMETER instead of EXCEPTION_PARAMETER. Until it's
                // fixed we manually inspect enclosing catch blocks.
                final TreePath path = getPath(node);
                Tree ct = TreeUtils.enclosingOfKind(path, Tree.Kind.CATCH);
                if (ct != null && elt.equals(TreeUtils.elementFromDeclaration(((CatchTree)ct).getParameter())))
                    type.addAnnotation(NONNULL);
            }

            annotateIfStatic(elt, type);
            return super.visitIdentifier(node, type);
        }
    }
}
