package checkers.util;

import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Name;

import checkers.types.QualifierHierarchy;

/**
 * Represents the type qualifier hierarchy of a type system.
 *
 * It uses a an internal graph to represent the hierarchy.
 *
 * In addition to simply representing qualifiers, this implementation allows
 * for two special qualifiers:
 * <ul>
 *  <li>Root qualifier: a qualifier that is a super qualifier of all other
 *      types in the hierarchy.  One doesn't need to declare all of its
 *      relations.</li>
 *  <li>Wildcard qualifier: a qualifier that matches any other type.  In other
 *      words is a sub and super of qualifiers.</li>
 * </li>
 */
@Deprecated
public class GraphQualifierHierarchy extends QualifierHierarchy {
    /** map: qualifier --> supertypes of the qualifier**/
    private Map<AnnotationMirror, Set<AnnotationMirror>> subtypes;

    /** the root of all the qualifiers **/
    private AnnotationMirror root;

    /** the bottom of all qualifiers **/
    private AnnotationMirror wildcard;
    private boolean hasWildcard = false;

    /**
     * Constructs an instance of {@code GraphQualifierHierarchy} with
     * no qualifier (i.e. {@code null}) being the root qualifier
     */
    public GraphQualifierHierarchy() {
        this(null);
    }

    /**
     * Initialize an instance with the provided annotation being the root
     * annotation in the hierarchy
     *
     * @param root  the type qualifier that is the supertype of all other qualifiers
     */
    public GraphQualifierHierarchy(AnnotationMirror root) {
        subtypes = AnnotationUtils.createAnnotationMap();
        setRootAnnotation(root);
    }

    /**
     * Add a type qualifier to the type hierarchy
     *
     * @param anno  a type qualifier
     * @return  true if is added to the hierarchy graph,
     *          false if it's already in the hierarchy graph
     */
    public boolean addTypeQualifier(AnnotationMirror anno) {
        if (subtypes.containsKey(anno))
            return false;
        subtypes.put(anno, AnnotationUtils.createAnnotationSet());
        return true;
    }

    /**
     * Adds a subtype relationship between two type qualifiers.
     * One does not need to declare subtyping relations to the root or
     * placeholder type qualifiers.
     *
     * @param subAnno   the sub type qualifier
     * @param superAnno the super type qualifier
     */
    public void addSubtype(AnnotationMirror subAnno, AnnotationMirror superAnno) {
        addTypeQualifier(subAnno);
        addTypeQualifier(superAnno);

        if (AnnotationUtils.areSame(subAnno, superAnno))
            return;

        subtypes.get(subAnno).add(superAnno);
        Set<AnnotationMirror> supSupers = subtypes.get(superAnno);

        if (supSupers != null) {
            for (AnnotationMirror supSuper : supSupers)
                subtypes.get(subAnno).add(supSuper);
        }
    }

    @Override
    public AnnotationMirror getRootAnnotation() {
        return root;
    }

    /**
     * Set the root type qualifier of the represented hierarchy
     * @param root  the root (the ultimate super) type qualifier
     */
    public void setRootAnnotation(AnnotationMirror root) {
        this.root = root;
        addTypeQualifier(root);
    }

    /**
     * Declare a wildcard annotation that matches all annotations
     * @param placeHolder   the place holder annotation
     */
    public void setWildcard(AnnotationMirror placeHolder) {
        this.wildcard = placeHolder;
        addTypeQualifier(placeHolder);
        this.hasWildcard = true;
    }

    @Override
    public boolean isSubtype(AnnotationMirror sub, AnnotationMirror sup) {
        if (AnnotationUtils.areSame(sup, this.root) || AnnotationUtils.areSame(sup, sub))
            return true;
        if (hasWildcard &&
                (AnnotationUtils.areSame(sup, this.wildcard) ||
                        AnnotationUtils.areSame(sub, wildcard)))
            return true;

        return subtypes.get(sub).contains(sup);
    }

    @Override
    public Set<Name> getTypeQualifiers() {
        Set<Name> names = new HashSet<Name>();
        for (AnnotationMirror anno: subtypes.keySet())
            names.add(AnnotationUtils.annotationName(anno));
        return Collections.unmodifiableSet(names);
    }

    @Override
    public boolean inConflict(AnnotationMirror a1, AnnotationMirror a2) {
        return subtypes.containsKey(a1) && subtypes.containsKey(a2);
    }

    @Override
    public AnnotationMirror leastUpperBound(AnnotationMirror a1,
            AnnotationMirror a2) {
        return this.root;
    }
}
